#include <unistd.h>
#include <time.h>
#include <X11/Xlib.h>
#include <fcntl.h>
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "plugin.h"
#include "pixmap.h"
#include "cpu.xpm"
#include "cpu_qnx.xpm"
#include "fan.xpm"
#include "pci.xpm"
#include "temperature.xpm"
#include "voltage.xpm"

static char **icons[] = {cpu_qnx_xpm,cpu_xpm,fan_xpm,pci_xpm,temperature_xpm,voltage_xpm};

#define MAX_ICONS (sizeof(icons)/sizeof(icons[0]))

typedef struct _tagRECT
{
  int left;
  int top;
  int right;
  int bottom;
} RECT,*LPRECT;

typedef enum {ST_TEMPERATURE,ST_FAN,ST_VOLTAGE} SENSORTYPE;

#define ICON_USE_DEFAULT -1
#define ICON_NO_ICON -2

typedef struct _tagSENSORMODULE
{
  MODULE base;

  /* member info */
  Display *pDisplay;
  Window root;
  Window panelWindow;
  int screen;
  LPFONT pFont;
  int textY;
  unsigned long backGround;
  int width;
  int height;
  unsigned long black;
  SENSORTYPE st;
  int requestedIcon;
  int convertFahrenheit;
  int fd;
  GC gc;
  PIXMAP icon;
  RECT arrowRect;
  char tooltip[128];
} SENSORMODULE,*LPSENSORMODULE;

#define OFFSET 12

#define ICON_SPACING 4

/* returns a boolean value, 0 on failure */
#define CTOI(a) (a - '0')

static int GetReading(LPSENSORMODULE pSensor,char *output)
{
  char buf[128];
  int len;
  double fah;
  if ((lseek(pSensor->fd,0,SEEK_SET) == 0) && (len=read(pSensor->fd,buf,sizeof(buf))) >= 4)
  {
    /* now remove the trailing carriage return if we get one */
    if (buf[len-1] == '\n')
    {
      buf[--len] = 0;
    }

    switch (pSensor->st)
    {
    case ST_TEMPERATURE:
      if (pSensor->convertFahrenheit)
      {
	fah = 32.0 + (9.0/5.0 * ((double)(CTOI(buf[0]) * 10 + CTOI(buf[1])) + CTOI(buf[2])/10.0));
	sprintf(output,"%3.1f",fah);
      }
      else			/* celcius */
      {
	output[0] = buf[0];
	output[1] = buf[1];
	output[2] = '.';
	output[3] = buf[2];
	output[4] = 0;
      }
      break;
    case ST_FAN:
      memcpy(output,buf,len);
      output[len] = 0;
      break;
    case ST_VOLTAGE:
      output[0] = buf[0];
      output[1] = '.';
      output[2] = buf[1];
      output[3] = buf[2];
      output[4] = 0;
      break;
    default:
      return 0;
    }
    return 1;
  }
  return 0;
}

static void LoadIcon(LPSENSORMODULE pSensor)
{
  char **icon = NULL;

  if ((pSensor->requestedIcon >= 0) && (pSensor->requestedIcon < MAX_ICONS))
  {
    icon = icons[pSensor->requestedIcon];
  }
  else if (pSensor->requestedIcon == ICON_USE_DEFAULT)
  {
    switch (pSensor->st)
    {
    case ST_TEMPERATURE:
      icon = temperature_xpm;
      break;
    case ST_FAN:
      icon = fan_xpm;
      break;
    case ST_VOLTAGE:
      icon = voltage_xpm;
      break;
    }
  }
  
  if (icon != NULL)
  {
    LoadPixmap(icon,&pSensor->icon);
  }
}

static void DrawSensor(LPSENSORMODULE pSensor,int drawBackground)
{
  /* get the time */
  char buf[128];
  int x;
  /* Erase the background */
  if (drawBackground)
  {
    XClearArea(pSensor->pDisplay,pSensor->base.panels[0].wnd,0,0,0,0,False);
  }

  if (!GetReading(pSensor,buf))
  {
    strcpy(buf,"Fail");
  }

  if (pSensor->icon.icon != None)
  {
    RenderPixmap(pSensor->pDisplay,pSensor->base.panels[0].wnd,pSensor->gc,pSensor->arrowRect.left,pSensor->arrowRect.top,&pSensor->icon);

    x = pSensor->icon.width + 2 * ICON_SPACING;
  }
  else
  {
    x = OFFSET/2;
  }
  pSensor->pFont->SetColor(pSensor->pFont,pSensor->black);
  pSensor->pFont->Render(pSensor->pFont,pSensor->base.panels[0].wnd,x,pSensor->textY,buf,strlen(buf));
}

static int LoadOptions(LPSENSORMODULE pSensor,LPCONFIGITEM pOptions)
{
  LPCONFIGITEM pItem;
  const char *pFile;
  /* set them to the defaults */
  if (pOptions && pOptions->pSublist)
  {
    for (pItem = (LPCONFIGITEM)pOptions->pSublist->list.pHead;pItem;pItem = (LPCONFIGITEM)pItem->node.pNext)
    {
      if (strcmp(pItem->name,"icon") == 0)
      {
	if (strcmp(pItem->value,"default") == 0)
	{
	  pSensor->requestedIcon = ICON_USE_DEFAULT;
	}
	else if (strcmp(pItem->value,"none") == 0)
	{
	  pSensor->requestedIcon = ICON_NO_ICON;
	}
	else			/* see if an index */
	{
	  pSensor->requestedIcon = atoi(pItem->value);
	}
      }
      else if (strcmp(pItem->name,"conversion") == 0)
      {
	if (strcmp(pItem->value,"F") == 0 || strcmp(pItem->value,"f") == 0)
	{
	  pSensor->convertFahrenheit = 1;
	}
      }
      else if (strcmp(pItem->name,"path") == 0)
      {
	/* now determine the type */
	if ((pFile = strrchr(pItem->value,'/')))
	{
	  ++pFile;
	}
	else 
	{
	  pFile = pItem->value;
	}
	/* now get the sensor type based on the file name */
	if (strncmp(pFile,"temp_input",10) == 0)
	{
	  pSensor->st = ST_TEMPERATURE;
	}
	else if (strncmp(pFile,"fan_input",9) == 0)
	{
	  pSensor->st = ST_FAN;
	}
	else if (strncmp(pFile,"in_input",8) == 0)
	{
	  pSensor->st = ST_VOLTAGE;
	}
	else
	{
	  /* unknown sensor, abort */
	  continue;
	}
	/* ok now try to open */
	pSensor->fd = open(pItem->value,O_RDONLY);
      }
    }
  } 
  return (pSensor->fd >= 0);
}

static const char *Sensor_GetTooltip(LPSENSORMODULE pSensor)
{
  char buf[128];
  
  if (!GetReading(pSensor,buf))
  {
    strcpy(buf,"Fail");
  }
  
  switch (pSensor->st)
  {
  case ST_TEMPERATURE:
    sprintf(pSensor->tooltip,"Temperature: %s %c",buf,pSensor->convertFahrenheit ? 'F' : 'C');
    break;
  case ST_FAN:
    sprintf(pSensor->tooltip,"Fan: %s RPMs",buf);
    break;
  case ST_VOLTAGE:
    sprintf(pSensor->tooltip,"Voltage: %s",buf);
    break;
  default:
    return NULL;
  }
  return pSensor->tooltip;
}

static void Sensor_Timeout(LPSENSORMODULE pSensor)
{
  DrawSensor(pSensor,1);
  XFlush(pSensor->pDisplay);
}

static void Sensor_EventHandler(LPSENSORMODULE pSensor,XEvent *pEvent)
{
  switch (pEvent->type)
  {
  case ButtonPress:
    switch (pEvent->xbutton.button)
    {
    case 1:
      DrawSensor(pSensor,1);
      XFlush(pSensor->pDisplay);
    }
    break;
  case Expose:
    if (pEvent->xexpose.count == 0)
    {
      DrawSensor(pSensor,0);
    }
    break;
  }
}

static void Sensor_Destroy(LPSENSORMODULE pSensor)
{
#ifdef DEBUG
  printf("sensors.so, Destroy()\n");
#endif
  if (pSensor->pFont)
  {
    pSensor->pFont->Release(pSensor->pFont);
  }
  
  if (pSensor->fd >= 0)
  {
    close(pSensor->fd);
  }

  if (pSensor->gc != None)
  {
    XFreeGC(pSensor->pDisplay,pSensor->gc);
  }

  FreePixmap(&pSensor->icon);

  free(pSensor);
}

static Window CreateWindow(LPSENSORMODULE pSensor,LPLIBRARYINIT pInit)
{
  XSetWindowAttributes att;
  Window ret;
  
  memset(&att,0,sizeof(att));
  
#ifdef DEBUG
  printf("Creating sensor window with a height of %i\n",pInit->height);
#endif
  pSensor->width = pInit->pFont->GetWidth(pInit->pFont,"999.9",5);
  if (pSensor->icon.icon != None)
  {
    pSensor->width += (pSensor->icon.width + 2 * ICON_SPACING) + OFFSET/2;
  }
  else
  {
    pSensor->width += OFFSET;
  }
  pSensor->height = pInit->height;
  
  /* now create the window */
  att.background_pixel = pInit->backgroundPixel;
  att.event_mask = pInit->baseInputMask | ButtonPressMask | ExposureMask;

  ret = XCreateWindow(pInit->pDisplay,pInit->panelWindow,0,0,pSensor->width,pInit->height,0,CopyFromParent,InputOutput,CopyFromParent,CWEventMask | CWBackPixel,&att);
  if (ret)
  {
    /* setup GC */
    pInit->pFont->GetCenter(pInit->pFont,"999.9",5,0,pInit->height,NULL,&pSensor->textY);
    /* add ref to the font */
    pSensor->pFont = pInit->pFont;
    pInit->pFont->AddRef(pInit->pFont);
  }
  return ret;
}

int Init(LPLIBRARYINIT pInit)
{
  XGCValues gcv;
  LPSENSORMODULE pSensor = (LPSENSORMODULE)malloc(sizeof(SENSORMODULE));
  if (pSensor)
  {
    memset(pSensor,0,sizeof(SENSORMODULE));

    pSensor->base.Destroy = (MODULE_DESTROY)Sensor_Destroy;
    pSensor->base.Timeout = (MODULE_TIMEOUT)Sensor_Timeout;
    pSensor->base.EventHandler = (MODULE_EVENTHANDLER)Sensor_EventHandler;
    pSensor->base.GetTooltip = (MODULE_GETTOOLTIP)Sensor_GetTooltip;
    pSensor->base.numPanels = 1;

    pSensor->pDisplay = pInit->pDisplay;
    pSensor->root     = pInit->root;
    pSensor->screen   = pInit->screen;
    pSensor->panelWindow = pInit->panelWindow;
    pSensor->backGround = pInit->backgroundPixel;
    pSensor->black  = BlackPixel(pInit->pDisplay,pInit->screen);
    pSensor->fd = -1;
    pSensor->requestedIcon = ICON_NO_ICON;
    
    /* load formats */
    if (LoadOptions(pSensor,pInit->pOptions) <= 0)
    {
      free(pSensor);
      return 0;
    }
    /* now load the icon */
    LoadIcon(pSensor);
    /* now create the window */
    if (!(pSensor->base.panels[0].wnd  = CreateWindow(pSensor,pInit)))
    {
      FreePixmap(&pSensor->icon);
      free(pSensor);
      return 0;
    }
    
    /* create the gc */
    if (pSensor->icon.icon != None)
    {
      pSensor->arrowRect.left = ICON_SPACING;
      pSensor->arrowRect.top = (pSensor->height - pSensor->icon.height)/2;
      pSensor->arrowRect.right = pSensor->arrowRect.left + pSensor->icon.width;
      pSensor->arrowRect.bottom = pSensor->arrowRect.top + pSensor->icon.height;
      gcv.graphics_exposures = False;
      gcv.clip_mask = pSensor->icon.mask;
      gcv.clip_x_origin = pSensor->arrowRect.left;
      gcv.clip_y_origin = pSensor->arrowRect.top;
      pSensor->gc = XCreateGC(pInit->pDisplay,pSensor->base.panels[0].wnd,GCGraphicsExposures | GCClipMask | GCClipXOrigin | GCClipYOrigin,&gcv);
    }
    pSensor->base.panels[0].name = "Sensor";
    /* set return value */
    pInit->pModule = (LPMODULE)pSensor;
  }
  return (pSensor != 0);
}

